(* Definition of basic Closures *)

(* close : (int -> int*pos) -> Games.play -> int -> int -> int -> Interaction.ask_action *)

open Games
open Interaction

type count = One | Two | Three | Four | Five | Even | Odd

let incr = function 
  | One -> Two
  | Two -> Three
  | Three -> Four
  | Four -> Five 
  | Five -> Even 
  | Even -> Odd
  | Odd -> Even

let visible_comp m last_component = 
  if 
    m.component = 0 || m.component = last_component 
  then
    true
  else 
    false


(* ar: arity of the identity. Only answer on the correct channels
 * f : channel n°i redirected on f i
 * interface: specifies the behavious of the copy/paste
 *            set it to n to get an identity n -> n
 * poly: The number of free variables in the interaction
 * curry: 0 if we are in left position. n if we are in right position and have n arguments*)


let id_close2 ar f interface poly curry p component dec last_component = 
  let initial m c = 
    match m.justifier with 
      | None -> true
      | Some k when k.component < c -> true
      | _ -> false 
  in

    match p.current with 
      | None -> 
	  Debug.debug "id_close2 : empty play";
	  Interaction.ANone
      | Some p_current when (initial p_current component) ->
	  let j = p_current 
	  and lab = 
	    match p_current.label with 
	      | Question i -> Question (f (i-dec))
	      | Answer k -> 
		  Debug.debug "id_close2 : initial answer !";
		  raise Out_of_domain
	  in 
	    (
	      match Games.add_move j lab p poly curry with 
		| Games.Left -> Interaction.ALeft
		| Games.Right -> Interaction.ARight
	    )
      | Some p_current -> 
	  (
	    match p_current.justifier with 
	      | None -> assert false
	      | Some one_above -> 
		  (
		    match one_above.previous with 
		      | None -> 
			  Debug.debug "id_close2 : initial player move !";
			  raise Out_of_domain
		      | Some two_above -> 
			  let label = 
			    if (initial two_above component) then 
			      (
				match p_current.label with 
				  | Question q -> Question (q+interface)
				  | Answer k -> Answer k
			      )
			    else
			      p_current.label
			  in
			    (
			      match Games.add_move two_above label p poly curry with 
				| Games.Left -> Interaction.ALeft
				| Games.Right -> Interaction.ARight
			    )
		  )
	  )
	  
let f_close2 f poly curry p component dec last_component =
  Debug.debug "I am f_close2 ! ";
  let initial m c =
    match m.justifier with
      | None -> true
      | Some k when k.component < c -> true
      | _ -> false
  in

    match p.current with
      | None ->
          Debug.debug "f_close2 : empty play";
          Interaction.ANone
      | Some p_current when (initial p_current component) ->
	  (
	    match Games.add_move p_current (Question poly) p poly curry with 
	      | Games.Left -> Interaction.ALeft
	      | Games.Right -> Interaction.ARight
	  )
      | Some p_current ->
	  (
	    match p_current.label with 
	      | Question _ -> 
		  Debug.debug "f_close2 : Encountered a non initial question";
		  Interaction.ANone
	      | Answer k -> 
		  (
		    match p_current.justifier with 
		      | None -> assert false
		      | Some j -> 
			  (
                            match j.previous with 
                                | None -> 
                                    Debug.debug "f_close2 : Missing anterior move";
                                    Interaction.ANone
                                | Some jp -> 
                                    (
                                      match Games.add_move jp (Answer (f k)) p poly curry with 
			                | Games.Left -> Interaction.ALeft
			                | Games.Right -> Interaction.ARight
                                    )
			  )
		  )
	  )

let f_bin_close f poly curry p component dec last_component =
  let initial m c =
    match m.justifier with
      | None -> true
      | Some k when k.component < c -> true
      | _ -> false
  in
    match p.current with
      | None -> 
        Interaction.ANone
      | Some p_current when (initial p_current component) ->
        (
	  match Games.add_move p_current (Question poly) p poly curry with
	    | Games.Left -> Interaction.ALeft
	    | Games.Right -> Interaction.ARight
	)
      | Some p_current ->
        (
	  match p_current.label with
	    | Question _ -> Interaction.ANone
	    | Answer k ->
	      (
	        match p_current.justifier with
		  | None -> assert false
		  | Some j -> 
		    (
		      match j.previous with 
		        | None -> assert false
			| Some jp -> 
			  (
			    match jp.label with
			      | Question q ->
			        (
				  match Games.add_move jp (Question (poly+1)) p poly curry with
				    | Games.Left -> Interaction.ALeft
				    | Games.Right -> Interaction.ARight
				)
			      | Answer k' -> 
			        (
				  match jp.justifier with 
				    | None -> assert false
				    | Some jpj -> 
				      (
				        match jpj.previous with 
					  | None -> assert false
					  | Some jpjp ->
					    (
					      match Games.add_move jpjp (Answer (f k' k)) p poly curry with
					        | Games.Left -> Interaction.ALeft
						| Games.Right -> Interaction.ARight
					    )
				      )
				)
			  )
		    )
	      )
	 )
	  

let c_close2 c poly curry p component dec last_component = 
  Debug.debug "I am c_close2";
  let initial m c =
    match m.justifier with
      | None -> true
      | Some k when k.component < c -> true
      | _ -> false
  in

    match p.current with 
      | None -> 
	  Debug.debug "c_close2 : empty play";
	  Interaction.ANone
      | Some p_current when (initial p_current component) -> 
	  (
	    match Games.add_move p_current (Answer c) p poly curry with 
	      | Games.Left -> Interaction.ALeft
	      | Games.Right -> Interaction.ARight
	  )
      | Some p_current -> 
	  Debug.debug "c_close2 : A non initial move, this should not be here.";
	  Interaction.ANone


let y_close poly curry p component dec last_component = 
  let initial m =
    match m.justifier with
      | None -> true
      | Some k when k.component < component -> true
      | _ -> false
  in

  Debug.debug "I am y_close";

  match p.current with 
    | None -> 
	Debug.debug "y_close : empty play";
	Interaction.ANone
    | Some p_current -> 
	(
	  match p_current.justifier with 
	    | None -> 
		(
		  match p_current.label with 
		    | Question i when i - dec = 0 -> 
			(
			  match Games.add_move p_current (Question poly) p poly curry with 
			    | Games.Left -> Interaction.ALeft
			    | Games.Right -> Interaction.ARight
			)
		    | _ -> Interaction.ANone
		)
	    | Some one_above when one_above.component < component -> 
		(
		  match p_current.label with 
		    | Question i when i - dec = 0 -> 
			(
			  match Games.add_move p_current (Question poly) p poly curry with 
			    | Games.Left -> Interaction.ALeft
			    | Games.Right -> Interaction.ARight
			)
		    | _ -> Interaction.ANone
		)
	    | Some one_above -> 
		
		(
		  match one_above.justifier with 
		    | None -> Interaction.ANone
		    | Some two_above when two_above.component < component -> Interaction.ANone
		    | Some two_above -> 
			(
			  match (initial two_above) with 
			    | true ->
				(
				  match one_above.label with 
				    | Answer k -> 
					(
					  match one_above.previous with 
					    | None -> assert false
					    | Some j -> 
						(
						  match Games.add_move j p_current.label p poly curry with 
						    | Games.Left -> Interaction.ALeft
						    | Games.Right -> Interaction.ARight
						)
					)
				    | Question q -> 
					let real_label = 
					  if one_above.component = 0 then 
					    q + curry
					  else 
					    q
					in 
					  (
					    match real_label with 
					      | q when q = poly -> 
						  (
						    match p_current.label with 
						      | Question q when q = 0 ->
							  (
							    match Games.add_move two_above (Question poly) p poly curry with 
							      | Games.Left -> Interaction.ALeft
							      | Games.Right -> Interaction.ARight
							  )
						      | Question q -> 
							  (
							    match one_above.previous with 
							      | None -> assert false
							      | Some j -> 
								  (
								    let lab = 
								      if j == two_above then 
									Question (q+poly)
								      else 
									Question (q-1)
								    in 
								      match Games.add_move j lab p poly curry with 
									| Games.Left -> Interaction.ALeft
									| Games.Right -> Interaction.ARight
								  )
							  )
						      | Answer k -> 
							  (
							    match one_above.previous with 
							      | None -> assert false
							      | Some j -> 
								  (
								    match Games.add_move j (Answer k) p poly curry with 
								      | Games.Left -> Interaction.ALeft 
								      | Games.Right -> Interaction.ARight
								  )
							  )
						  )
					      | _ -> 
						  (
						    match one_above.previous with 
						      | None -> assert false
						      | Some j -> 
							  (
							    match Games.add_move j p_current.label p poly curry with 
							      | Games.Left -> Interaction.ALeft 
							      | Games.Right -> Interaction.ARight
							  )
						  )
					  )
				)
				  
			    | false -> 
				(
				  match one_above.previous with 
				    | None -> assert false
				    | Some j -> 
					(
					  match Games.add_move j p_current.label p poly curry with 
					    | Games.Left -> Interaction.ALeft 
					    | Games.Right -> Interaction.ARight
					)
				)
			)
		)
	)

	  
let if_close2 poly curry p component dec last_component = 
  Debug.debug "I am if_close2";
  let initial m =
    match m.justifier with
      | None -> true
      | Some k when k.component < component -> true
      | _ -> false
  in

    match p.current with 
      | None -> 
	  Debug.debug "if_close2 : empty play !";
	  Interaction.ANone
      | Some p_current -> 
	  (
	    match (initial p_current) with 
	      | true -> 
		  (
		    match Games.add_move p_current (Question poly) p poly curry with 
		      | Games.Left -> Interaction.ALeft
		      | Games.Right -> Interaction.ARight
		  )
	      | false -> 
		  (
		    match p_current.justifier with 
		      | None -> assert false
		      | Some one_above -> 
			  (
			    match one_above.previous with 
			      | None -> failwith "Interaction.ANone2"
			      | Some two_above ->
				  (
				    match (initial two_above) with 
				      | true -> 
					  (
					    match one_above.label with 
					      | Question q -> 
						  let real_label = 
						    if one_above.component = 0 then 
						      q + curry 
						    else 
						      q 
						  in 
						    if real_label=poly then 
						      (
							match p_current.label with 
							  | Question _ -> failwith "Interaction.ANone"
							  | Answer 0 -> 
							      (
								match Games.add_move two_above (Question (poly+2)) p poly curry with 
								  | Games.Left -> Interaction.ALeft 
								  | Games.Right -> Interaction.ARight
							      )
							  | Answer _ -> 
							      (
								match Games.add_move two_above (Question (poly+1)) p poly curry with 
								  | Games.Left -> Interaction.ALeft
								  | Games.Right -> Interaction.ARight
							      )
						      )
						    else
						      (
							match p_current.label with 
							  | Question q -> 
							      (
								match Games.add_move two_above (Question (q+poly+3)) p poly curry with 
								  | Games.Left -> Interaction.ALeft
								  | Games.Right -> Interaction.ARight
							      )
							  | Answer k -> 
							      (
								match Games.add_move two_above (Answer k) p poly curry with 
								  | Games.Left -> Interaction.ALeft
								  | Games.Right -> Interaction.ARight
							      )
						      )
					      | Answer k -> 
						  (
						    match one_above.previous with 
						      | None -> assert false
						      | Some j -> 
							  (
							    match Games.add_move j p_current.label p poly curry with 
							      | Games.Left -> Interaction.ALeft
							      | Games.Right -> Interaction.ARight
							  )
						  )
					  )
				      | false -> 
					  (
					    match two_above.justifier with 
					      | None -> assert false
					      | Some jpj ->
					      (
					        match jpj.previous with
						  | None -> assert false
						  | Some jpjp ->
					      let lab = (
						  match p_current.label with 
						    | Question q -> Question (q+3+poly)
						    | l -> l) in
					        match (initial jpjp) with 
						  | true -> 
						    (
						      match Games.add_move jpjp lab p poly curry with 
						        | Games.Left -> Interaction.ALeft
						        | Games.Right -> Interaction.ARight
						    )
						  | false -> 
						    (
						      match Games.add_move two_above p_current.label p poly curry with 
						        | Games.Left -> Interaction.ALeft
						        | Games.Right -> Interaction.ARight
					  	    )
					       )
					   )
					  )
				  )
			  )
		  )
	  

let bot_close poly curry p component dec last_component = Interaction.ANone

let id ar f i= 
  {
    Interaction.c_arity = ar;
    Interaction.close = id_close2 ar f i
  }	      

let bot = 
  {
    Interaction.c_arity = 1;
    Interaction.close = bot_close
  }

let succ = 
  { 
    Interaction.c_arity = 1;
    Interaction.close = f_close2 (fun n -> n + 1 )
  }
      
let zero = 
  {
    Interaction.c_arity = 1;
    Interaction.close = c_close2 0
  }


let integer k = 
  {
    Interaction.c_arity = 1;
    Interaction.close = c_close2 k
  }


let ifthenelse = 
  {
    Interaction.c_arity = 1;
    Interaction.close = if_close2
  }

let iszero = 
  {
    Interaction.c_arity = 1;
    Interaction.close = f_close2 (fun n -> if n = 0 then 1 else 0)
  }

let pred = 
  { 
    Interaction.c_arity = 1;
    Interaction.close = f_close2 (function | 0 -> 0 | n -> n - 1)
  }

let y =
  {
    Interaction.c_arity = 1;
    Interaction.close = y_close
  }

let eq =
  {
    Interaction.c_arity = 1;
    Interaction.close = f_bin_close (fun n m -> if n=m then 1 else 0)
  }

let plus =
  {
    Interaction.c_arity = 1;
    Interaction.close = f_bin_close (fun n m -> n + m)
  }

let times =
  {
    Interaction.c_arity = 1;
    Interaction.close = f_bin_close (fun n m -> n*m)
  }

let minus =
  {
    Interaction.c_arity = 1;
    Interaction.close = f_bin_close (fun n m -> n - m)
  }

let bnot =
  {
    Interaction.c_arity = 1;
    Interaction.close = f_close2 (function | 0 -> 1 | _ -> 0)
  }

let band =
  {
    Interaction.c_arity = 1;
    Interaction.close = f_bin_close (fun n m -> match n,m with | 0,_ -> 0 | _,0 -> 0 | _,_ -> 1)
  }

let bor =
  { 
    Interaction.c_arity = 1;
    Interaction.close = f_bin_close (fun n m -> match n,m with | 1,_ -> 1 | _,1 -> 1 | _,_ -> 0)
  }

