open Games
open Interaction
open Closures


type lambda = Var of int | Abstr of int*lambda | App of lambda*lambda | AppFoc of lambda*lambda | If | S | O | IsZero | Pred 
		| Fixpoint | Integer of int | Equal | Plus | Times | Minus | Not | And | Or
 
let rec string_of_lambda = function 
        | Var i -> "Var "^(string_of_int i)
        | Abstr(i,l) -> "Abstr("^(string_of_int i)^","^(string_of_lambda l)^")"
        | App(l1,l2) -> "App("^(string_of_lambda l1)^","^(string_of_lambda l2)^")"
	| AppFoc(l1,l2) -> "AppFoc("^(string_of_lambda l1)^","^(string_of_lambda l2)^")"
        | S -> "S"
	| If -> "If"
	| IsZero -> "IsZero"
	| Pred -> "Pred"
        | Fixpoint -> "Fixpoint"
        | O -> "O"
        | Integer k -> "Integer "^(string_of_int k)
	| Equal -> "="
	| Plus -> "+"
	| Times -> "*"
	| Minus -> "-"
	| Not -> "~"
	| And -> "&"
	| Or -> "|"


exception Unbound_variable

let check_hnfa = 
	let rec get_hoc = function
		| Var i as h -> h
		| Abstr(i,l) -> get_hoc l
		| App(l1,l2) -> get_hoc l1
		| AppFoc(l1,l2) -> get_hoc l1
		| h -> h
	in function 
		| Var i -> true
		| Abstr(i,l) -> (match get_hoc l with 
					| Var j when i=j -> true
					| _ -> false
				)
		| App(l1,l2) -> false
		| AppFoc(l1,l2) -> false
		| _ -> true




let rec int_worker (n:int) f = function
  | App(App(App(App(l1,l2),l3),l4),l5) -> 
      let left_strat = 
	match n with 
	  | 0 -> 
	      {
		agent = 
		  Tensor(
		    Tensor(int_worker n f l2,
			   int_worker n f l3),
		    Tensor(int_worker n f l4,
			   int_worker n f l5));
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 Tensor(Tensor(int_worker n f l2,
				       int_worker n f l3),
				Tensor(int_worker n f l4,
				      int_worker n f l5)));
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker n f l1;
	  curry = 4
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }
  | App(App(App(l1,l2),l3),l4) -> 
      let left_strat = 
	match n with 
	  | 0 -> 
	      {
		agent = 
		  Tensor(
		    Tensor(int_worker n f l2,
			   int_worker n f l3),
		    int_worker n f l4);
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 Tensor(Tensor(int_worker n f l2,
				       int_worker n f l3),
				int_worker n f l4));
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker n f l1;
	  curry = 3
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }
  | App(App(l1,l2),l3) -> 
      let left_strat = 
	match n with 
	  | 0 -> 
	      {
		agent = 
		  Tensor(int_worker n f l2,
			 int_worker n f l3);
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 Tensor(int_worker n f l2,
				int_worker n f l3));
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker n f l1;
	  curry = 2
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }

  | App(l1,l2) -> 
      let left_strat = 	
	match n with 
	  | 0 -> 
	      {
		agent = int_worker 0 f l2;
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 int_worker n f l2);
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker n f l1;
	  curry = 1
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }

  | AppFoc(l1,l2) -> 
      let left_strat = 	
	match n with 
	  | 0 -> 
	      {
		agent = int_worker 0 f l2;
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 int_worker n f l2);
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker n f l1;
	  curry = 1
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }
	  
  | Abstr(i,m) -> int_worker (n+1) (fun x -> if x = i then n else f x) m
      
  | Var i -> Closure (id 1 (let j = f i in fun _ -> j) n)

  | S -> Closure succ

  | O -> Closure zero

  | Integer k -> Closure (integer k)

  | If -> Closure ifthenelse

  | IsZero -> Closure iszero

  | Pred -> Closure pred

  | Fixpoint -> Closure y

  | Equal -> Closure eq

  | Plus -> Closure plus

  | Times -> Closure times

  | Minus -> Closure minus

  | Not -> Closure bnot

  | And -> Closure band

  | Or -> Closure bor
      

let init_interp m i = match m with 
  | App(m1,m2) -> 
      let f = fun k -> failwith ("interpretation : variable "^(string_of_int k)^" is unknown...") in 
      let left_strat = 	
	{ 
	  agent = 
	    int_worker 0 f m2;
	  curry = 0
	}
      and right_strat = 
	{
	  agent = int_worker 0 f m1;
	  curry = 1
	} 
      and p = new_play (Some (new_move i)) in
	  { 
	    arity = 1;
	    play = p;
	    strategies = [|right_strat; left_strat|];
	    status = InTau 0;
	    poly = 0
	  }
  | _ -> failwith "init_interp: no interaction at toplevel"
	  
(*	
*NEW INTERPRETATION FUNCTION :
*	* Takes a lambda-term as argument, returns an agent with a pointer to the active cut
*	* If there is no active cut, take as active the outmost cut
*	* If there is _no_ cut, create one with the bottom strategy
*)

let smart_intr n f l = 
let rec int_worker2 n f = function
(*  | App(App(App(App(l1,l2),l3),l4),l5) -> 
      let left_strat = 
	match n with 
	  | 0 ->
	      {
		agent = 
		  Tensor(
		    Tensor(int_worker2 n f l2,
			   int_worker2 n f l3),
		    Tensor(int_worker2 n f l4,
			   int_worker2 n f l5));
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 Tensor(Tensor(int_worker2 n f l2,
				       int_worker2 n f l3),
				Tensor(int_worker2 n f l4,
				       int_worker2 n f l5)));
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker2 n f l1;
	  curry = 4
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  } 
  | App(App(App(l1,l2),l3),l4) -> 
      let left_strat = 
	match n with 
	  | 0 -> 
	      {
		agent = 
		  Tensor(
		    Tensor(int_worker2 n f l2,
			   int_worker2 n f l3),
		    int_worker2 n f l4);
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 Tensor(Tensor(int_worker2 n f l2,
				       int_worker2 n f l3),
				int_worker2 n f l4));
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker2 n f l1;
	  curry = 3
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }
  | App(App(l1,l2),l3) -> 
      let left_strat = 
	match n with 
	  | 0 -> 
	      {
		agent = 
		  Tensor(int_worker2 n f l2,
			 int_worker2 n f l3);
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 Tensor(int_worker2 n f l2,
				int_worker2 n f l3));
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker2 n f l1;
	  curry = 2
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }*)
  | App(l1,l2) -> 
      let left_strat = 	
	match n with 
	  | 0 -> 
	      {
		agent = int_worker2 0 f l2;
		curry = 0
	      }
	  | _ -> 
	      { 
		agent = 
		  Tensor(Closure (id n (fun x -> x) n),
			 int_worker2 n f l2);
		curry = 0
	      }
      and right_strat = 
	{
	  agent = int_worker2 n f l1;
	  curry = 1
	} in 
	Interaction 
	  { 
	    arity = 1;
	    play = {current = None};
	    strategies = [|right_strat; left_strat|];
	    status = Ready_i;
	    poly = n
	  }

  | AppFoc(l1,l2) -> 
      let left_strat =
        match n with
          | 0 ->
              {
                agent = int_worker2 0 f l2;
                curry = 0
              }
          | _ ->
              {
                agent =
                  Tensor(Closure (id n (fun x -> x) n),
                         int_worker2 n f l2);
                curry = 0
              }
      and right_strat =
        {
          agent = int_worker2 n f l1;
          curry = 1
        } in 
	Interaction
          {
            arity = 1;
            play = {current = None};
            strategies = [|right_strat; left_strat|];
            status = Ready_i;
            poly = n
          } 

  | Abstr(i,m) -> int_worker2 (n+1) (fun x -> if x = i then n else f x) m
      
  | Var i -> Closure (id 1 (let j = f i in fun _ -> j) n)

  | S -> Closure succ

  | O -> Closure zero

  | Integer k -> Closure (integer k)

  | If -> Closure ifthenelse

  | IsZero -> Closure iszero

  | Pred -> Closure pred

  | Fixpoint -> Closure y

  | Equal -> Closure eq

  | Plus -> Closure plus

  | Times -> Closure times

  | Minus -> Closure minus

  | Not -> Closure bnot

  | And -> Closure band

  | Or -> Closure bor
in int_worker2 n f l

let init_interp2 m i = 
	let f = fun k -> failwith ("interpretation : variable "^(string_of_int k)^" is unknown...") in
	let ag = smart_intr 0 f m in
	 match ag with 
	 	| Interaction i2 -> 
			let p = new_play (Some (new_move i)) in
				{
				 arity = i2.arity;
				 play = p;
				 strategies = i2.strategies;
				 status = InTau 0;
				 poly = i2.poly
				} 
		| _ ->
			Debug.debug "Warning : Toplevel agent is not an interaction. Creating fake interaction with bottom.";
			let left_strat = 
			  {
			    agent = Closure Closures.bot;
			    curry = 0
			  }
			and right_strat = 
			  {
			    agent = ag;
			    curry = 1
			  }
			and p = new_play (Some (new_move i)) in 
                                {
                                 arity = 1;
                                 play = p;
                                 strategies = [| right_strat ; left_strat |];
                                 status = InTau 0;
                                 poly = 0
                                }


let addr_of_lambda lam = 
	let rec addr_worker n = function
		| AppFoc(l1,l2) -> Some []
		| App(l1,l2) -> 
		(
		match addr_worker n l1 with 
			| Some addr -> Some ('R'::addr)
			| None -> 
			(
			match addr_worker n l2 with 
				| Some addr -> if n=0 then 
						Some ('L'::addr)
					       else 
					        Some ('L'::'U'::addr)
				| None -> None
			)
		)
		| Abstr(i,l) -> 
		(
		match addr_worker (n+1) l with
			| Some addr as ad -> ad
			| None -> None
		)
		| _ -> None
in match (addr_worker 0 lam) with 
	| Some cl -> Common.string_of_char_list cl
	| None -> ""

