
open Lambda
open Interaction

exception Computation_finished of Games.play*int
exception Computation_failure of Games.play*int

let formula p =
    let h = Dump.play_to_hashtbl p in
	Dump.formula h



let output_play p = 
  try
    let n = Games.length p in	
    (if  n > 577 then raise Dump.Play_too_big else ());
    let h = Dump.play_to_hashtbl p in
    let v = Dump.hashtbl_to_vector h in
    let s = Dump.vector_to_string v in
    let out_c = open_out "out.strid" in 
      output_string out_c "
matrix {\n";
      output_string out_c s;
      output_string out_c "
}";
      close_out out_c;
      assert ((Sys.command (Printf.sprintf "strid --pdf out.strid &> /dev/null")) = 0)
  with 
    | e -> raise e


let output_play_colors p = 
  try
    let h = Dump.play_to_hashtbl_colors p in
    let v = Dump.hashtbl_to_vector h in
    let s = Dump.vector_to_string v in
    let out_c = open_out "out.strid" in 
      output_string out_c "
matrix {\n";
      output_string out_c s;
      output_string out_c "
}";
      close_out out_c;
      assert ((Sys.command (Printf.sprintf "strid --pdf out.strid &> /dev/null")) = 0)
  with 
    | e -> raise e

let intr_of_address addr i = 
  let n = String.length addr in
  let rec aux ag intr = function 
    | i when i = n -> 
	(
	  match ag with 
	    | Closure _ -> 
		Debug.warning "Address is a closure. Keeping the last found interaction.";
		Debug.debug ("Active Interaction : "^(string_of_agent (Interaction intr)));
		intr
	    | Tensor(t1,t2) -> 
		Debug.warning "Address is a tensor. Keeping the last found interaction.";
		Debug.debug ("Active Interaction : "^(string_of_agent (Interaction intr)));
		intr
	    | Interaction i -> 
		Debug.debug "Address valid.";
		Debug.debug ("Active Interaction : "^(string_of_agent (Interaction i)));
		i
	)
    | k -> 
	(
	  match ag with 
	    | Closure _ -> 
		Debug.warning "Address is invalid. Keeping the last found interaction.";
		Debug.debug ("Active Interaction : "^(string_of_agent (Interaction intr)));
		intr;
	    | Tensor(ag1,ag2) -> 
		if addr.[k] = 'U' then 
		  aux ag2 intr (k+1)
		else if addr.[k] = 'D' then 
		  aux ag1 intr (k+1)
		else
		  (
		    Debug.warning "Address is invalid. Keeping the last found interaction.";
		    Debug.debug ("Active Interaction : "^(string_of_agent (Interaction intr)));
		    intr
		  )
	    | Interaction i -> 
		if addr.[k] = 'L' then 
		  aux (i.strategies.(1).agent) i (k+1)
		else if addr.[k] = 'R' then 
		  aux (i.strategies.(0).agent) i (k+1)
		else 
		  (
		    Debug.warning "Address is invalid. Keeping the last found interaction.";
		    Debug.debug ("Active Interaction : "^(string_of_agent (Interaction intr)));
		    intr
		  )
	)
  in aux (Interaction i) i 0


let toplevel i_out addr = 
  let total = ref 0 in 
  let focus = ref (intr_of_address addr i_out) in 
    while true do
      print_string ((string_of_int !total)^"> ");
      let c = read_line() in 
	match c with 
	  | "" -> ()
	  | "help" -> print_endline "Available commands are : print, print_focus, output, run [k], focus [addr], exit, colors"
	  | "print" -> print_endline ("Toplevel agent"^(string_of_agent (Interaction i_out)))
	  | "print_focus" -> print_endline ("Focused agent"^(string_of_agent (Interaction (!focus))))
	  | "formula" -> print_endline((formula ((!focus).play)))
	  | "output" -> 
	      (
		try 
		  output_play ((!focus).play)
		with 
		  (*| failure s ->  Debug.debug s*)
		  | Dump.Play_too_big -> 
                      print_endline "Play is too big to be outputed";
		      flush stdout
                  | e -> 
                      print_endline "Output failed.";flush stdout;
                      raise e
	      )
	  | "colors" -> 
	      (
		try 
		  output_play_colors ((!focus).play)
		with 
		  (*| failure s ->  Debug.debug s*)
		  | Dump.Play_too_big -> 
                      print_endline "Play is too big to be outputed";
		      flush stdout
                  | e -> 
                      print_endline "Output failed.";flush stdout;
                      raise e
	      )
	  | "views" -> 
	  	print_endline (Misc.string_lengths ((!focus).play))
          | "exit" -> exit 0
	  | s -> 
	      if String.sub s 0 3 = "run" then 
		let n = String.length s in 
		let k = try
		  int_of_string (String.sub s 4 (n-4)) 
		with 
		  | _ -> 1000000000 in 
		  for j = 1 to k do
		    match (total:=!total + 1);(Interaction.step (Interaction i_out)) with 
		      | InTau _ as a -> Debug.debug (string_of_status a)
		      | Ready_i | Ready_o as a ->
			  Debug.debug (string_of_status a);
			  raise (Computation_finished ((!focus).play,!total))
		      | _ -> raise (Computation_finished ((!focus).play,!total))
		  done
	      else if String.sub s 0 5 = "focus" then 
		let n = String.length s in 
		  if n > 6 then 
		    let addr = String.sub s 6 (n-6) in 
		      focus := intr_of_address addr i_out
		  else focus := intr_of_address "" i_out
	      else
		print_endline ("Unknown command : "^s)
    done
		
	
    


let _ = 
  try 
    print_endline "T ::= λx.T | TT | T@T | x | S | k | if | iszero | pred | Y | = | + | - | * | not | & | '|'  ";
    print_string "Entrez le terme : ";
    flush stdout;
    let lexbuf = Lexing.from_channel stdin in
    let arg_lambda = Parser.main Lexer.token lexbuf in
      Debug.debug_p (string_of_lambda arg_lambda);
      let i_out = init_interp2 arg_lambda 0 and
          init_addr = addr_of_lambda arg_lambda in
	try 
	  toplevel i_out init_addr
	with 
	  | Computation_finished (p,i) -> 
              print_endline ("Computation finished in "^(string_of_int i)^" moves.");
	      output_play p
	  | e -> raise e
  with End_of_file -> exit 0
    | e -> raise e
    
