open Hashtbl
open Games
  
(* flof_play : takes a play, a function 'a -> 'a -> polarity -> 'b, and applies this function on each node of the play with justifier as left argument and previous as right argument *)
(* First, we define a fold on the play which returns a hashtbl with
   all the computed information on moves. Then, we use it to define
   a classic fold function on plays *)

let fold_hash p f a_prev a_point =
  let n = Games.length p in
  let h = Hashtbl.create n in
  let rec aux m =
    match m.previous with 
      | None -> 
          let res = f a_point a_prev Opponent m in
	  Hashtbl.add h m (res,Opponent);
	  (res,Opponent)
      | Some prev ->
	    (match m.justifier with 
	      | None -> 
		    let (a,_) =
		    try
		      Hashtbl.find h prev
		    with
		      | Not_found -> aux prev in
		  let res = f a_point a Opponent m in
		  Hashtbl.add h m (res,Opponent);
                  (res,Opponent)
	      | Some just -> 
		  let (a1,_) = try
		    Hashtbl.find h just
		  with | Not_found -> aux just in
		  let (a2,pol) = try Hashtbl.find h prev 
		    with | Not_found -> aux prev in
		    let res = f a1 a2 (opp pol) m in
		    Hashtbl.add h m (res,opp pol);
		    (res,opp pol)
		)
in match p.current with 
	| None -> h
	| Some m -> ignore (aux m); h


let flof_play p f a_prev a_point =(*initially a typo of fold_play, but it was funny *)
   match p.current with 
    | None -> a_prev
    | Some m -> 
   let h = fold_hash p (fun r1 r2 pol _ -> f r1 r2 pol) a_prev a_point in
	fst (Hashtbl.find h m)
			    
let p_o_lengths p = 
 let l =
  flof_play p 
    (fun l1 l2 pol -> 
       let (j_p_length,j_o_length) = List.hd l1
       and (p_p_length,p_o_length) = List.hd l2 in
	 match pol with 
	   | Opponent -> 
	       (j_p_length + 1,p_o_length + 1)::l2
	   | Player -> 
	       (p_p_length + 1,j_o_length + 1)::l2
    )
    [(0,0)] [(0,0)]
 in List.tl (List.rev l)

(* We define here a function that takes a move, and returns how
far it points in its view. If the move is not visible, we return
0. It is sound, since a move can never point to itself. For the
moment, 0 will be used as well when the move has no justifier *)

let locate_pointer m =
  let rec aux n m' m_goal pol = 
    if m' == m_goal then n else match pol with
      | Player -> 
        (
	  match m'.previous with 
	    | None -> 0
	    | Some m'' -> 
	    aux (n+1) m'' m_goal Opponent
	)
      | Opponent -> 
        (
          match m'.justifier with
	    | None -> 0
	    | Some m'' -> 
	    aux (n+1) m'' m_goal Player
	)
in match m.justifier with 
	| None -> 0
	| Some j -> aux 0 m j Player

let compute_views p = 
  fold_hash p 
    (fun (o1,p1) (o2,p2) pol m ->
	  let n = locate_pointer m in
      match pol with 
        | Player -> 
	    ((1,m.label)::o1,(n,m.label)::p2)
	| Opponent ->
	    ((n,m.label)::o2,(1,m.label)::p1)
    ) ([],[]) ([],[])
		


let show_list show l =
  let rec aux = function 
    | [] -> "]"
    | t::q -> (show t)^";"^(aux q)
  in ("["^(aux l))

let show = function (a,b) -> "("^(string_of_int a)^","^(string_of_int b)^")"

let string_lengths p = show_list show (p_o_lengths p)


